
# Created by Amin Rahmani, Ph.D., P.E. - July 14, 2019
# Copyright © 2019 Amin Rahmani. All rights reserved.

#========================================================================================
#                               ****   Notes  ****
# This code performs STATIC lateral soil-beam interaction analysis using p-y springs. The code can be used for both soil-pile and soil-wall interaction evaluations. 
# The "uniaxialMaterial ENT" was used in this model to enable the end-user to define different backnone curves on left and right sides of a retaining wall.
# You need to makes changes on the lines that the comment "User-Defined" appears. Please go through all the command lines before running it.
# Units are Imperial (lbf, ft, sec).  
# Outputs will be saved in a folder titled "Outputs".
# p-y springs below the ground level will be generated at every #zSize (typically 2 ft) depth of the proposed pile.
#========================================================================================

wipe all

set startT [clock seconds]

##### This will create an input file for GID program if you want to check the final geometry.
set meshFile [open soilmesh.flavia.msh w]
##########################################################################################
	

#### Definition of Global Directions
####################################
### Dir Z: Starts at ground surface and extends downward below ground
### Dir X: To the left side
### Dir Y: Perpendicular to the plane of the paper

#######################################################################
#### Input parameters used for creating the geometry of soil-pile model 
#######################################################################
set gravX  0.0000
set gravY  0.0000
set gravZ  32.2				         ;# gravity in direction Z. The value would be +9.81 if metric units are used.

#### Pile length                                  
set LP       60.0                        ;# User-defined --- Total pile length (from top to tip of the pile)
set LP_emb   50.0			             ;# User-defined --- embedded pile length below the ground surface.\
							             ;# \in soil-wall analysis this length would be the embedded depth of pile below the excavation bottom
                                         ;# \in front face of the wall.

#### Pile element size
set zSize    2.0                         ;# User-Defined 


if { int($LP_emb/$zSize) == $LP_emb/$zSize} {
	set numZele_emb  [expr int($LP_emb/$zSize)]           ;# number of elements in z  direction    direction
     } else {
	set numZele_emb  [expr int($LP_emb/$zSize)+1]           ;# number of elements in z  direction    direction
	}

if { $LP_emb == $LP } {
	set numZele_above_grade  0           ;# number of elements in z  direction    direction
     } else {
	   if { int(($LP-$LP_emb)/$zSize) == ($LP-$LP_emb)/$zSize } {
	      set numZele_above_grade  [expr int(($LP-$LP_emb)/$zSize)]           ;# number of elements in z  direction    direction
         } else {
	      set numZele_above_grade  [expr int(($LP-$LP_emb)/$zSize)+1]           ;# number of elements in z  direction    direction
	   }
     }

########################################################
#### Introduce Pile And Springs into the Numerical Model
########################################################

##############################################
#### Define Pile Nodes
##############################################

model basic -ndm 3 -ndf 6

#### For GID Program Use
puts $meshFile "MESH beamcolumn dimension 3 ElemType linear Nnode 2"
puts $meshFile "Coordinates"
#######################################################################


### Define Pile Nodes Below Ground Level
for {set k 1} { $k <= [expr $numZele_emb+1] } {incr k 1} {
     
      set xdim 0.0
      set ydim 0.0

      if { int($LP_emb/$zSize) == $LP_emb/$zSize} {
         set   zdim [expr ($k-1)*$zSize]
      } else {  
         if { $k <= $numZele_emb } {
             set   zdim [expr ($k-1)*$zSize]
         } else {
             set   zdim $LP_emb
         }
	}
	
     node $k $xdim $ydim $zdim 
     
     #### Boundary Condition 
     fix  $k 0 1 1 1 0 1
     #######################
     
     #### For GID Program Use
      puts $meshFile "$k $xdim $ydim $zdim"
     ###########################################
}

### Define Pile Nodes Above Ground Level
if { $numZele_above_grade == 0 }  {
    ### Do nothing
   } else {
   for {set k 1 } { $k <= $numZele_above_grade } {incr k 1} {
  
      set xdim 0.0
      set ydim 0.0

      if { int(($LP-$LP_emb)/$zSize) == ($LP-$LP_emb)/$zSize} {
         set   zdim [expr -$k*$zSize]
      } else {  
         if { $k <= [expr $numZele_above_grade-1] } {
             set   zdim [expr -$k*$zSize]
         } else {
             set   zdim [expr -($LP-$LP_emb)]
         }
	}
	
     node [expr $numZele_emb+1+$k] $xdim $ydim $zdim 
     
     #### Boundary Condition 
     fix  [expr $numZele_emb+1+$k] 0 1 1 1 0 1
     ########################

	#### For GID Program Use
      puts $meshFile "[expr $numZele_emb+1+$k] $xdim $ydim $zdim"
	###########################################
  }
}


##############################################
#### Define Spring Nodes Below Ground Level
##############################################

model basic -ndm 3 -ndf 3


if { int($LP_emb/$zSize) == $LP_emb/$zSize} {

        set Num_Springs [expr $numZele_emb+1]  
    } else {  
	
        set Num_Springs [expr $numZele_emb] 
    }
		 

for {set k 1} { $k <= $Num_Springs } {incr k 1} {
     
      set xdim 0.0
      set ydim 0.0
      set   zdim [expr ($k-1)*$zSize]
	
     node [expr 1000+$k] $xdim $ydim $zdim 
     node [expr 2000+$k] $xdim $ydim $zdim 
     node [expr 3000+$k] $xdim $ydim $zdim 
	
	#### Boundary Condition 
	fix [expr 2000+$k] 1 1 1
	fix [expr 3000+$k] 1 1 1

	equalDOF  [expr $k] [expr 1000+$k]  1 2 3
	
	#######################

	#### For GID Program Use
      puts $meshFile "[expr 1000+$k] $xdim $ydim $zdim"
	puts $meshFile "[expr 2000+$k] $xdim $ydim $zdim"
	puts $meshFile "[expr 3000+$k] $xdim $ydim $zdim"
	###########################################
}


model basic -ndm 3 -ndf 6

##############################################
#### Define Spring Nodes At The Pile Head
##############################################

# #### User Defined --- Comment out this block if there is no spring
# set xdim 0.0
# set ydim 0.0
# set zdim [expr -($LP-$LP_emb)]
	
# node 6000 $xdim $ydim $zdim 
# node 7000 $xdim $ydim $zdim 

# #### Boundary Condition 
# fix  7000 1 1 1 1 1 1

# equalDOF  6000 [expr $numZele_emb+$numZele_above_grade+1] 1 2 3 4 5 6

##############################################
#### Pile Head Fixity
##############################################

##### Fixed-Head
##fix [expr $numZele_emb+$numZele_above_grade+1] 0 1 1 1 1 1 ; # User-Defined

##### Free-Head
fix [expr $numZele_emb+$numZele_above_grade+1] 0 1 1 1 0 1 ; # User-Defined

#### For GID Program Use
puts $meshFile "end coordinates"
puts $meshFile "Elements"
#########################

##############################################
#### Define Pile Cross-Section & Material
##############################################

set Pile_sec_mat_tag 1 

### Pile section dimensions
set D    3.0		;# User-Defined --- Pile Cross section diameter
set t    0.042		;# User-Defined --- Pile Cross section wall thickness
set A    [expr 0.25*3.1416*pow($D,2)-0.25*3.1416*pow($D-2.0*$t,2)] 	      ;# User-Defined
set Iz   [expr 0.25*3.1416*pow($D/2.0,4)-0.25*3.1416*pow($D/2.0-$t,4)]		;# User-Defined 
set Iy   [expr 0.25*3.1416*pow($D/2.0,4)-0.25*3.1416*pow($D/2.0-$t,4)]		;# User-Defined 
set J    [expr $Iz+$Iy]

### Section Type 1: Elastic Section --- User-Defined
set E     4176.0e+6	;# User-Defined --- Pile material elastic modulus
set G    [expr $E/2.0/(1+0.20)]
section Elastic $Pile_sec_mat_tag $E $A $Iz $Iy $G $J

### Section Type 2: Nonliner Section --- User-Defined
#set D              1.52 						  ; #Use-defined --- Pier diameter
#set E_concrete     20.0e+6						  ; #Use-defined --- concrete Young's modulus 
#set Ebar           200.0e+6						  ; #Use-defined --- bar Young's modulus 
#set Apier          [expr 0.25*3.1416*pow($pierD,2)]      ; #Use-defined --- pier cross-sectional area 
#set rhopier        2.45					        ; #Use-defined --- pier material density
#set Fypier         455.0e+3						   ; #Use-defined --- pier yield strength
#set areaFiber      0.002581						   ; #Use-defined 
#set massDens_pier  [expr $rhopier*$Apier]
#
#set fc -34473.8							       ; #Use-defined --- concrete strength
#set epsc0  [expr 2.0*$fc/$E_concrete]
#uniaxialMaterial Steel01 8001 $Fypier $Ebar 0.0                          ; #Use-defined
#uniaxialMaterial Steel02 8002 $Fy $Ebar 0.0 18 0.925 0.15                ; #Use-defined 
#uniaxialMaterial Concrete01  8003  -34473.8  -0.004   -21000.0  -0.014   ; #Use-defined
#uniaxialMaterial Concrete01  8013  -27600.0  -0.002   0.0  -0.008        ; #Use-defined
#
#section Fiber $Pile_sec_mat_tag { 
#  patch circ  8003 50 50 0.0 0.0 0. [expr $pierD/2.0-0.074] 0.0 360.0; #Use-defined
#  patch circ  8013 50 50 0.0 0.0 [expr $pierD/2.0-0.074] [expr $pierD/2.0] 0.0 360.0; #Use-defined
#  layer circ  8001 18 $areaFiber 0.0 0.0 0.686 0.0 360.0 ; #Use-defined
#  }


#######################################################
#### Define Pile Elements
#######################################################

geomTransf Linear 101 0 1 0

set numIntgrPts 5

### Define Pile Element Below Ground Level
for { set k 1 } { $k <= [expr $numZele_emb] } { incr k 1 } {

      element dispBeamColumn $k $k [expr $k+1] $numIntgrPts $Pile_sec_mat_tag  101   
      puts $meshFile "$k $k [expr $k+1] 10"
     }

### Define Pile Element Above Ground Level
if { $numZele_above_grade == 0 }  {
    ### Do nothing
   } else {
    for { set k 1 } { $k <= [expr $numZele_above_grade] } { incr k 1 } {

	 if { $k == 1 } {
  	       element dispBeamColumn [expr $numZele_emb+$k] 1 [expr $numZele_emb+1+$k] $numIntgrPts $Pile_sec_mat_tag  101   
  		 puts $meshFile "[expr $numZele_emb+$k] 1 [expr $numZele_emb+1+$k] 10"
	 } else {
  	 	 element dispBeamColumn [expr $numZele_emb+$k] [expr $numZele_emb+$k] [expr $numZele_emb+1+$k] $numIntgrPts $Pile_sec_mat_tag  101   
  		 puts $meshFile "[expr $numZele_emb+$k] 1 [expr $numZele_emb+1+$k] 10"

      }
   }
}

#############################
#### Define Spring Materials
#############################

model basic -ndm 3 -ndf 3  

##### Dummy Rigid Spring With Zero Tension
##########################################

set E_Spring 1.0e+22  ;# Rigid Summy Spring

uniaxialMaterial ENT 1000 $E_Spring

##### P-Y Springs Left Side of Beam
###########################################

#### User-Defined --- Imported from LPile Program --- Must start from 2001 (ground surface) ---- Number of elements should be the same as "$Num_Springs"
#### At the ground surface, do not use 0.0 force values. Use very small number such as 0.0001 lbf instead. For example "uniaxialMaterial MultiLinear	1001	0.02	0.0001	0.04	0.0001	0.08	0.0001	0.25	0.0001"
uniaxialMaterial MultiLinear	1001	0.02	0.0001	0.05	0.0001	0.10	0.0001	0.270833333	0.0001
uniaxialMaterial MultiLinear	1002	0.02	3923	0.05	5377	0.10	5598	0.270833333	5597.920139
uniaxialMaterial MultiLinear	1003	0.02	7977	0.05	11064	0.10	11612	0.270833333	11611.62039
uniaxialMaterial MultiLinear	1004	0.02	11426	0.05	15200	0.10	15698	0.270833333	15698.39805
uniaxialMaterial MultiLinear	1005	0.02	13527	0.05	15957	0.10	16100	0.270833333	16100.41376
uniaxialMaterial MultiLinear	1006	0.02	29728	0.05	31899	0.10	42994	0.270833333	42993.79511
uniaxialMaterial MultiLinear	1007	0.02	58011	0.05	62499	0.10	82527	0.270833333	82526.86196
uniaxialMaterial MultiLinear	1008	0.02	74906	0.05	84375	0.10	96281	0.270833333	96281.33744
uniaxialMaterial MultiLinear	1009	0.02	92965	0.05	109128	0.10	110025	0.270833333	110025.0984
uniaxialMaterial MultiLinear	1010	0.02	111235	0.05	136510	0.10	138589	0.270833333	138589.1717
uniaxialMaterial MultiLinear	1011	0.02	128739	0.05	166233	0.10	170379	0.270833333	170379.0124
uniaxialMaterial MultiLinear	1012	0.02	147257	0.05	198279	0.10	205383	0.270833333	205383.3682
uniaxialMaterial MultiLinear	1013	0.02	166789	0.05	231777	0.10	243587	0.270833333	243586.7862
uniaxialMaterial MultiLinear	1014	0.02	187024	0.05	266779	0.10	284966	0.270833333	284965.704
uniaxialMaterial MultiLinear	1015	0.02	204049	0.05	304175	0.10	329459	0.270833333	329458.9062
uniaxialMaterial MultiLinear	1016	0.02	220859	0.05	341242	0.10	377034	0.270833333	377033.8792
uniaxialMaterial MultiLinear	1017	0.02	237964	0.05	378447	0.10	427654	0.270833333	427653.8222
uniaxialMaterial MultiLinear	1018	0.02	255365	0.05	417450	0.10	481137	0.270833333	481137.0883
uniaxialMaterial MultiLinear	1019	0.02	273062	0.05	458250	0.10	537513	0.270833333	537513.3165
uniaxialMaterial MultiLinear	1020	0.02	291055	0.05	500847	0.10	596823	0.270833333	596823.3453
uniaxialMaterial MultiLinear	1021	0.02	309344	0.05	540243	0.10	658217	0.270833333	658217.4493
uniaxialMaterial MultiLinear	1022	0.02	327930	0.05	577936	0.10	722599	0.270833333	722599.3883
uniaxialMaterial MultiLinear	1023	0.02	346810.77	0.05	616642.04	0.10	789806.50	0.27	789806.50
uniaxialMaterial MultiLinear	1024	0.02	365988.01	0.05	656362.49	0.10	858021.68	0.27	858021.68
uniaxialMaterial MultiLinear	1025	0.02	385461.29	0.05	697096.92	0.10	928656.37	0.27	928656.37
uniaxialMaterial MultiLinear	1026	0.02	202615.29	0.05	369422.67	0.10	501030.94	0.27	501030.94


##### P-Y Springs Right Side of Beam
###########################################

#### User-Defined --- Imported from LPile Program --- Must start from 2001 (ground surface)  ---- Number of elements should be the same as "$Num_Springs"
#### At the ground surface, do not use 0.0 force values. Use very small number such as 0.0001 lbf instead. For example "uniaxialMaterial MultiLinear	1001	0.02	0.0001	0.04	0.0001	0.08	0.0001	0.25	0.0001"
uniaxialMaterial MultiLinear	2001	0.02	0.0001	0.05	0.0001	0.10	0.0001	0.27	0.0001
uniaxialMaterial MultiLinear	2002	0.02	3923.32	0.05	5377.45	0.10	5597.92	0.27	5597.92
uniaxialMaterial MultiLinear	2003	0.02	7977.46	0.05	11063.67	0.10	11611.62	0.27	11611.62
uniaxialMaterial MultiLinear	2004	0.02	11425.57	0.05	15200.34	0.10	15698.40	0.27	15698.40
uniaxialMaterial MultiLinear	2005	0.02	13526.59	0.05	15957.33	0.10	16100.41	0.27	16100.41
uniaxialMaterial MultiLinear	2006	0.02	29727.57	0.05	31899.20	0.10	42993.80	0.27	42993.80
uniaxialMaterial MultiLinear	2007	0.02	58010.60	0.05	62498.84	0.10	82526.86	0.27	82526.86
uniaxialMaterial MultiLinear	2008	0.02	74905.84	0.05	84375.43	0.10	96281.34	0.27	96281.34
uniaxialMaterial MultiLinear	2009	0.02	92965.16	0.05	109128.25	0.10	110025.10	0.27	110025.10
uniaxialMaterial MultiLinear	2010	0.02	111234.95	0.05	136510.48	0.10	138589.17	0.27	138589.17
uniaxialMaterial MultiLinear	2011	0.02	128738.91	0.05	166232.73	0.10	170379.01	0.27	170379.01
uniaxialMaterial MultiLinear	2012	0.02	147256.85	0.05	198278.97	0.10	205383.37	0.27	205383.37
uniaxialMaterial MultiLinear	2013	0.02	166788.78	0.05	231776.72	0.10	243586.79	0.27	243586.79
uniaxialMaterial MultiLinear	2014	0.02	187024.01	0.05	266779.07	0.10	284965.70	0.27	284965.70
uniaxialMaterial MultiLinear	2015	0.02	204049.44	0.05	304174.57	0.10	329458.91	0.27	329458.91
uniaxialMaterial MultiLinear	2016	0.02	220858.56	0.05	341241.87	0.10	377033.88	0.27	377033.88
uniaxialMaterial MultiLinear	2017	0.02	237963.67	0.05	378447.50	0.10	427653.82	0.27	427653.82
uniaxialMaterial MultiLinear	2018	0.02	255364.80	0.05	417450.22	0.10	481137.09	0.27	481137.09
uniaxialMaterial MultiLinear	2019	0.02	273061.96	0.05	458250.05	0.10	537513.32	0.27	537513.32
uniaxialMaterial MultiLinear	2020	0.02	291055.14	0.05	500846.96	0.10	596823.35	0.27	596823.35
uniaxialMaterial MultiLinear	2021	0.02	309344.32	0.05	540243.05	0.10	658217.45	0.27	658217.45
uniaxialMaterial MultiLinear	2022	0.02	327929.54	0.05	577935.56	0.10	722599.39	0.27	722599.39
uniaxialMaterial MultiLinear	2023	0.02	346810.77	0.05	616642.04	0.10	789806.50	0.27	789806.50
uniaxialMaterial MultiLinear	2024	0.02	365988.01	0.05	656362.49	0.10	858021.68	0.27	858021.68
uniaxialMaterial MultiLinear	2025	0.02	385461.29	0.05	697096.92	0.10	928656.37	0.27	928656.37
uniaxialMaterial MultiLinear	2026	0.02	202615.29	0.05	369422.67	0.10	501030.94	0.27	501030.94


##### Springs in Series on the Left and Right Sides of Pile/Wall
################################################################

for { set k 1 } { $k <= [expr $Num_Springs] } { incr k 1 } {

     uniaxialMaterial Series [expr 3000+$k] 1000 [expr 1000+$k]
     uniaxialMaterial Series [expr 4000+$k] 1000 [expr 2000+$k]
}


#############################
#### Define Spring Elements
#############################

for { set k 1 } { $k <= [expr $Num_Springs] } { incr k 1 } {
	 
	 ###Springs on the leftside of the beam
     element zeroLength [expr 5000+$k]  [expr 1000+$k]  [expr 3000+$k]  -mat [expr 3000+$k] -dir 1 -orient  1 0 0 0  1 0
	 	      
	 ###Springs on the rightside of the beam
     element zeroLength [expr 4000+$k]  [expr 1000+$k]  [expr 2000+$k]  -mat [expr 4000+$k] -dir 1 -orient -1 0 0 0 -1 0
}


##### Translational and Rotational Springs at the Pile Head
################################################################

model basic -ndm 3 -ndf 6

# #### User Defined --- Comment out this block if there is no spring
# set K_X    1100.0E-4 ; #Spring stiffness in direction X
# set K_teta 1250.0E-4 ; #Spring rotational stiffness about Y axis

# uniaxialMaterial Elastic 6001 $K_X
# uniaxialMaterial Elastic 6002 $K_teta

# element zeroLength 8000  7000  6000  -mat 6001 6002 -dir 1 5 -orient  1 0 0 0 1 0


#### For GID Program Use
puts $meshFile "end elements"
close $meshFile

#########################################################################
#### Record Data
#########################################################################

set dataDir Outputs
file mkdir $dataDir

### Pile Recorders
eval "recorder Node -file $dataDir/disppile_below_ground.out           -time -nodeRange 1 [expr $numZele_emb+1] -dof 1 2 3 4 5 6 disp"
eval "recorder Element -file $dataDir/pileforcesGlob_below_ground.out  -time -eleRange  1 $numZele_emb  globalForce"

eval "recorder Node -file $dataDir/disppile_above_ground.out           -time -nodeRange [expr $numZele_emb+2] [expr $numZele_emb+1+$numZele_above_grade] -dof 1 2 3 4 5 6 disp"
eval "recorder Element -file $dataDir/pileforcesGlob_above_ground.out  -time -eleRange  [expr $numZele_emb+1] [expr $numZele_emb+$numZele_above_grade]  globalForce"

eval "recorder Element -file $dataDir/Soilreaction_left.out  -time -eleRange  5001 [expr 5000+$Num_Springs]  force"
eval "recorder Element -file $dataDir/Soilreaction_right.out  -time -eleRange  4001 [expr 4000+$Num_Springs]  force"

#########################################################################
#### Apply Load on Pile Head or Unifrom Load
#########################################################################

loadConst -time 0.0
wipeAnalysis

pattern Plain 2000 "Linear" {
	
	######## Pile Head Displacement and Rotation
    #sp [expr $numZele_emb+$numZele_above_grade+1] 1 0.25 ;# User-Defined
    
	######## Pile Head Load	
	#load [expr $numZele_emb+$numZele_above_grade+1] 200000.0 0.0 0.0 0.0 0.0 0.0 ;# User-Defined
	
	######## Distributed Pressure	
	for { set k 1 } { $k <= [expr $numZele_emb-1] } { incr k 1 } {
    		load [expr $k+1] [expr 26880.0+($k-1)*2880.0] 0.0 0.0 0.0 0.0 0.0 ;# User-Defined
     }
	#### load at pile tip 
	load [expr $numZele_emb+1] 47640.0 0.0 0.0 0.0 0.0 0.0 ;# User-Defined
	#### Load at pile top element			
	eleLoad -ele 1  -type -beamPoint 12360 0.0 0.16 ;# User-Defined
}

#########################################################################
#### Pile Static Analysis 
#########################################################################

set NumIncr 500 ;# User-Defined - Number of steps

integrator LoadControl [expr 1.0/$NumIncr] 
###integrator LoadControl 1 1 1 1
#numberer RCM
numberer Plain
#constraints Transformation
constraints Penalty 1.e14 1.e14
test NormDispIncr 1.0e-4 100 1
algorithm KrylovNewton
#algorithm Newton
#system BandGeneral
#system SparseGeneral
#system Mumps
system SparseSYM
analysis Static

 for {set numIncr 1} {$numIncr <= $NumIncr } {incr numIncr 1} {
       puts "##### Pushover Analysis step : $numIncr #####";
       analyze 1 
     }

puts " ## End of Analysis!! ## "


set endT [clock seconds]
set ElapsedTime [expr $endT-$startT]
set ElapsedHours [expr int($ElapsedTime/3600.)]
set ElapsedMins [expr int($ElapsedTime/60.-$ElapsedHours*60.)]
set ElapsedSecs [expr int($ElapsedTime-$ElapsedHours*3600-$ElapsedMins*60.)]
puts "Execution time: $ElapsedHours hours and $ElapsedMins minutes and $ElapsedSecs seconds !!"

########################################
#### End of the TCL Code
########################################

wipe



